import datetime
import logging
import subprocess
import sys
from .config import loadControl

log = logging.getLogger()


def release():
    """ Release a package
    """
    version = loadControl()['version']
    hash = subprocess.check_output(['git', 'rev-parse', '--short', 'HEAD']).strip()
    date = datetime.datetime.now().strftime('%d/%m/%Y, %H:%M:%S')
    branch = 'release/v{version}-{hash}'.format(version=version, hash=hash)
    message = '"Release v{version} ({hash}) on {date}"'.format(version=version, hash=hash, date=date)
    message_noci = message[:-1] + ' [skip ci]"'

    def git_clean():
        exit_code = subprocess.call('git diff-index --quiet HEAD', shell=True)
        if exit_code == 0:
            return True
        else:
            subprocess.call('git diff-index HEAD', shell=True)
            return False

    def git_action(message, command, error='failed'):
        log.info(message)
        log.debug('command: {}'.format(command))
        exit_code = subprocess.call(command, shell=True)
        log.debug('exit code = {}'.format(exit_code))
        if exit_code != 0:
            log.error('failed')
            sys.exit(1)

    if not git_clean():
        log.critical('Cannot release with a dirty repository, please commit your changes')
        sys.exit(1)

    git_action('check if merge is needed', 'git pull --no-edit')
    # if not git_clean() what to do?

    git_action('retrieve remote tags', 'git pull --tags')

    git_action('create release branch', 'git checkout -b {branch} develop'.format(branch=branch))

    git_action('move to master branch', 'git checkout master')
    git_action('pull if needed', 'git pull --no-edit')
    git_action('merge release branch to master', 'git merge --no-ff {branch} --no-commit'.format(branch=branch))
    if not git_clean():
        git_action('commit the merged result', 'git commit -m {message} -a'.format(message=message))
    git_action('tag the merged result', 'git tag -a v{version}-{hash} -m {message}'.format(
        version=version,
        hash=hash,
        message=message))

    git_action('move to develop branch', 'git checkout develop')
    git_action('merge release branch to develop', 'git merge --no-ff {branch} --no-commit'.format(branch=branch))
    if not git_clean():
        git_action('commit the merged result', 'git commit -m {message} -a'.format(message=message_noci))

    git_action('delete release branch', 'git branch -d {branch}'.format(branch=branch))

    git_action('push all commits', 'git push --all')
    git_action('push all tags', 'git push --tags')



import logging
import plumbum

log = logging.getLogger()


def upload(server, user, keyfile, filename):
    """ Upload *filename* to *user*@*server*:/var/www/wapt/ using SSH key *keyfile*
    """

    remote = plumbum.machines.SshMachine(server, user=user, keyfile=keyfile)
    fro = plumbum.local.path(filename)
    to = remote.path('/var/www/wapt/' + filename)
    plumbum.path.utils.copy(fro, to)
    to.chown('wapt', 'www-data')
    to.chmod(0644)

    remote['wapt-scanpackages /var/www/wapt/']



import argparse
import logging
import os
import sys
from .vers import __version__
from .check import isPackage
from .version import versionChecker
from .validate import validate
from .create import create
from .build import build
from .release import release
from .hash import sha256sum
from .annotate import annotate
from .upload import upload

log = logging.getLogger()


def cli():
    parser = argparse.ArgumentParser(prog='wapt', description='WAPT packaging utility.')
    subparsers = parser.add_subparsers(
        title='commands',
        dest='command')

    subparsers.add_parser(
        'version',
        help='display cli version')

    parser_check = subparsers.add_parser(
        'check-version',
        help='check if new upstream package exists')

    parser_check.add_argument(
        '--force',
        help='force download for big files',
        action='store_true')

    parser_check.add_argument(
        '--chat',
        help='send results to chat',
        action='store_true')

    parser_check.add_argument(
        '--badge',
        help='generate badge',
        action='store_true')

    subparsers.add_parser(
        'validate',
        help='validate package')

    parser_new = subparsers.add_parser(
        'create',
        help='create a new package')

    parser_new.add_argument(
        '--name',
        help='name of the new package to create',
        type=str)

    parser_build = subparsers.add_parser(
        'build',
        help='build package from current folder')

    parser_build.add_argument(
        '--key',
        help='key path to sign package',
        type=str)

    parser_build.add_argument(
        '--release',
        help='release number of package',
        type=int)

    parser_build.add_argument(
        '--maturity',
        help='maturity of package, overrides WAPT/control content',
        type=str)

    parser_upload = subparsers.add_parser(
        'upload',
        help='upload package')

    parser_upload.add_argument(
        '--server',
        help='WAPT server',
        type=str)

    parser_upload.add_argument(
        '--user',
        help='User to connect server with',
        type=str)

    parser_upload.add_argument(
        '--keyfile',
        help='SSH private key',
        type=str)

    parser_upload.add_argument(
        '--filename',
        help='File to upload',
        type=str)

    subparsers.add_parser(
        'release',
        help='release package to production')

    subparsers.add_parser(
        'setup',
        help='initialize folder after git clone')

    parser_checksum = subparsers.add_parser(
        'checksum',
        help='compute checksum')

    parser_checksum.add_argument(
        '--file',
        help='file to be analyzed',
        type=str)

    # parser_annotate = subparsers.add_parser(
    subparsers.add_parser(
        'annotate',
        help='add annotation to grafana')

    parser.add_argument('--verbose', help='verbose output', action='store_true')
    parser.add_argument('--silent', help='no output except critical messages', action='store_true')

    args = parser.parse_args()

    logging_config = logging.StreamHandler(sys.stdout)
    logging_config.setFormatter(logging.Formatter('[%(asctime)s - %(levelname)8s] %(message)s'))
    log.addHandler(logging_config)

    if args.silent:
        log.setLevel(logging.CRITICAL)
    elif args.verbose:
        log.setLevel(logging.DEBUG)
    else:
        log.setLevel(logging.INFO)

    log.info('WAPT cli v{}'.format(__version__))

    if args.command == 'version':
        sys.exit(0)

    if args.command == 'check-version':
        if not isPackage():
            sys.exit(1)

        mismatch = versionChecker(chat=args.chat, badge=args.badge, force=args.force)
        if mismatch:
            if 'CI' in os.environ:
                sys.exit(0)
            else:
                sys.exit(1)
        else:
            sys.exit(0)

    if args.command == 'validate':
        if not isPackage():
            sys.exit(1)

        if not validate():
            log.error('not a valid package')
            sys.exit(1)
        else:
            log.info('valid package')
            sys.exit(0)

    if args.command == 'create':
        create(args.name)
        sys.exit(0)

    if args.command == 'build':
        build(key=args.key, release=args.release, maturity=args.maturity)
        sys.exit(0)

    if args.command == 'upload':
        upload(server=args.server, user=args.user, keyfile=args.keyfile, filename=args.filename)
        sys.exit(0)

    if args.command == 'release':
        if not isPackage():
            sys.exit(1)

        release()
        sys.exit(0)

    if args.command == 'setup':
        if sys.platform == 'win32':
            command = 'git config --local core.fileMode false'
            log.debug(command)
            os.system(command)
        command = 'git flow init --defaults --force'
        log.debug(command)
        os.system(command)
        if sys.platform != 'win32':
            command = 'pre-commit install'
            log.debug(command)
            os.system(command)
        sys.exit(0)

    if args.command == 'checksum':
        checksum = sha256sum(args.file)
        log.info('{file} checksum is {checksum}'.format(file=args.file, checksum=checksum))
        sys.exit(0)

    if args.command == 'annotate':
        annotate()
        log.info('Adding annotation')
        sys.exit(0)



import anybadge
import bs4
import datetime
import json
import logging
import pefile
import re
import requests
import tempfile
import os
from packaging import version
from pkg_resources import resource_string
from .config import loadControl, loadVersionCheck
from .download import save_url_to_file

log = logging.getLogger()


def getExeVersion(file):
    pe = pefile.PE(file)

    if 'VS_FIXEDFILEINFO' not in pe.__dict__:
        version = '0.0.0.0'
    elif not pe.VS_FIXEDFILEINFO:
        version = '0.0.0.0'
    else:
        verinfo = pe.VS_FIXEDFILEINFO[0]

        version = '{:d}.{:d}.{:d}.{:d}'.format(
            verinfo.FileVersionMS >> 16,
            verinfo.FileVersionMS & 0xFFFF,
            verinfo.FileVersionLS >> 16,
            verinfo.FileVersionLS & 0xFFFF)

    pe.close()
    return version


def latestVersion(force=False):
    """ Extract latest version defined by version-check.json

    Parameters
    ----------
    force: bool
        force download of big files to check version

    Returns
    -------
    version: string
        version extracted from web page
    """
    config = loadVersionCheck()

    if 'external_check' not in config:
        config['external_check'] = True

    control = loadControl()

    if not config['external_check']:
        return control['version'].split('-')[0]

    url_info = requests.head(config['url_version'], allow_redirects=True)
    # log.debug(url_info)
    if url_info.headers['Content-Type'] != 'application/octet-stream':
        content = requests.get(config['url_version']).text.strip()

        if config['html'] is True:
            soup = bs4.BeautifulSoup(content, 'html.parser')
            if 'index' in config:
                index = config['index']
            else:
                index = 0

            latest_version = soup.select(config['selector'])[index].contents[0].strip()

            # log.debug('index = {}'.format(index))
            # selected = soup.select(config['selector'])
            # log.debug('selected = {}'.format(selected))
            # log.debug('selected[{}] = {}'.format(index, selected[index]))
            # log.debug('selected[{}].contents[0] = {}'.format(index, selected[index].contents[0]))
            # latest_version = selected[index].contents[0].strip()
        else:
            latest_version = content

        if 'cleaners' in config and len(config['cleaners']) > 0:
            for cleaner in config['cleaners']:
                # log.debug('applying cleaner: {}'.format(cleaner))
                if 'grep' not in cleaner:
                    latest_version = re.sub(
                            cleaner['pattern'],
                            cleaner['rewrite'],
                            latest_version,
                            flags=re.DOTALL).strip()
                else:
                    result = []
                    for line in latest_version.split('\n'):
                        if re.match(cleaner['pattern'], line):
                            result.append(re.sub(cleaner['pattern'], cleaner['rewrite'], line))

                    if 'sort' in cleaner and cleaner['sort'] is True:
                        result.sort(key=lambda s: map(int, s.split('.')))
                        result = result[::-1]  # reverse list

                    latest_version = result[0]

    else:
        # log.debug(url_info.headers)
        if int(url_info.headers['Content-Length']) > 5000000:
            log.debug('file is {}Mb, more than 5Mb'.format(int(url_info.headers['Content-Length']) / 1000000))
        if int(url_info.headers['Content-Length']) < 5000000 or datetime.date.today().weekday() == 3 or force:
            file = tempfile.mkstemp()
            os.close(file[0])
            save_url_to_file(config['url_version'], file[1])
            latest_version = getExeVersion(file[1])
            os.unlink(file[1])
        else:
            log.debug('file size greater than 5Mb, returning current version')
            # log.debug(control['version'])
            latest_version = control['version'].split('-')[0]

    return latest_version.replace('\n', ' ').replace('\r', '')


def versionChecker(chat=False, badge=False, force=False):
    """ Compare latest version defined by version-check.json, versus WAPT/control one

    Parameters
    ----------
    chat: bool
        send results to chat
    badge: bool
        generate upstream badge

    Returns
    -------
    results: bool
        no upgrade needed, the current version is greater or equal to to latest version
    """
    control = loadControl()
    log.info('Current {} version: {}'.format(control['name'], control['version']))
    v_current = version.parse(control['version'])
    log.debug('v_current ' + str(v_current))

    latest_version = latestVersion(force)
    log.info('Latest {} version: {}'.format(control['name'], latest_version))
    v_latest = version.parse(latest_version)
    log.debug('v_latest ' + str(v_latest))

    if v_latest > v_current:
        log.info('New version available, please upgrade package')

        if chat:
            log.debug('Sending message to Chat webhook ...')
            if 'CHAT_WEBHOOK_URL' in os.environ:
                payload = resource_string('wapttools.data', 'chat_message.json')
                payload = payload.replace('{package}', control['name'])
                payload = payload.replace('{old_version}', control['version'])
                payload = payload.replace('{new_version}', latest_version)
                payload = payload.replace('{homepage}', control['homepage'])

                # Compact JSON string
                payload = json.dumps(json.loads(payload), separators=(',', ':'))

                requests.session().post(
                    os.environ['CHAT_WEBHOOK_URL'],
                    data=payload,
                    headers={'Content-Type': 'application/json; charset=UTF-8'}
                )
            else:
                log.warning('CHAT_WEBHOOK_URL environment variable not defined, unable to send chat message')
            log.debug('... done.')

        if badge:
            log.debug('generating badge ...')
            badge = anybadge.Badge(label='upstream', value='new', default_color='red')
            badge.write_badge('upstream.svg')

        return True

    if badge:
        log.debug('generating badge ...')
        badge = anybadge.Badge(label='upstream', value='ok', default_color='green')
        badge.write_badge('upstream.svg')

    return False



import logging
import sys

log = logging.getLogger()


def commands(downloader=None):
    """ Default main function for WAPT setup.py
    """

    logging_config = logging.StreamHandler(sys.stdout)
    logging_config.setFormatter(logging.Formatter('[%(asctime)s - %(levelname)8s] %(message)s'))
    log.addHandler(logging_config)
    log.setLevel(logging.INFO)

    if 'debug' in sys.argv:
        log.setLevel(logging.DEBUG)
        log.debug('DEBUG mode enabled')

    if downloader is not None:
        downloader()



import logging
import os
try:
    import psutil
except ImportError:
    psutil = None
import tempfile


def logger(package, userspace=False):
    logger = logging.getLogger(package)
    logger.setLevel(logging.DEBUG)

    if userspace:
        if psutil:
            current_user = psutil.Process().username().lower()
        else:
            current_user = os.getlogin()
        # Remove Windows domain if needed
        if '\\' in current_user:
            current_user = current_user.split('\\')[1]
        logfile = os.path.join(tempfile.gettempdir(), 'WAPT', current_user, package + '.log')
    else:
        logfile = os.path.join(os.environ['SYSTEMROOT'], 'WAPT', 'logs', package + '.log')

    if not os.path.isdir(os.path.dirname(logfile)):
        os.makedirs(os.path.dirname(logfile))

    handler = logging.FileHandler(logfile)
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(funcName)s: %(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    return logger



import glob
import logging
import os
import requests
import shutil
import sys
from .hash import sha256sum

log = logging.getLogger()


def copy_tree(src, dst, symlinks=False, ignore=None):
    if not os.path.exists(dst):
        os.makedirs(dst)
    for item in os.listdir(src):
        s = os.path.join(src, item)
        d = os.path.join(dst, item)
        if os.path.isdir(s):
            copy_tree(s, d, symlinks, ignore)
        else:
            if not os.path.exists(d) or os.stat(s).st_mtime - os.stat(d).st_mtime > 1:
                shutil.copy2(s, d)


def save_url_to_file(url, pathname):
    if url.startswith('http'):
        try:
            r = requests.get(url, stream=True)
            if not r.ok:
                log.critical('failed to download {} from {}, 404 not found'.format(os.path.basename(pathname), url))
                sys.exit(1)

            with open(pathname, 'wb') as f:
                for chunk in r.iter_content(1024):
                    f.write(chunk)
            del r
        except Exception as e:
            log.critical('failed to download {} from {}, {}'.format(os.path.basename(pathname), url, e))
            sys.exit(1)
    elif url.startswith('file'):
        from_path = url[7:]
        if os.path.isdir(from_path):
            copy_tree(from_path, pathname)
        else:
            shutil.copy2(from_path, pathname)
    else:
        log.critical('failed to download {} from {}, invalid url'.format(os.path.basename(pathname), url))
        sys.exit(1)


def download(url, pathname, clean=True, checksum=None):
    """ Download the file at a given url and save it to pathname.

    The files existing in the same folder and with the same extension are removed if clean=True.

    Parameters
    ----------
    url: string
        url of the file to be dowloaded
    pathname: string
        pathname of the file to be saved
    clean: boolean
        cleanup needed
    """
    extension = os.path.splitext(pathname)[1]

    dirname = os.path.dirname(pathname)

    if clean:
        pattern = os.path.join(dirname, '*{}'.format(extension))
        log.debug('cleaning files {} except {}'.format(pattern, pathname))
        for file in glob.glob(pattern):
            if file != pathname:
                log.debug('removing {}'.format(file))
                os.remove(file)

    log.info('downloading {} from {}'.format(os.path.basename(pathname), url))

    if not os.path.exists(pathname):
        save_url_to_file(url, pathname)
    else:
        if checksum:
            computed = sha256sum(pathname)
            if computed != checksum:
                log.info('{file} checksum mismatch, downloading'.format(file=pathname))
                os.remove(pathname)
                save_url_to_file(url, pathname)
        else:
            log.debug('{} already exists'.format(pathname))

    if checksum:
        computed = sha256sum(pathname)
        if computed != checksum:
            os.remove(pathname)
            log.critical('{file} checksum mismatch, downloaded file deleted'.format(file=pathname))
            sys.exit(1)



import logging
import os
import requests
import time
from .check import isPackage
from .config import loadControl

log = logging.getLogger()


def annotate():
    """ Generate annotation

    Parameters
    ----------
    """
    if not isPackage():
        log.error('current folder does not contain a WAPT package')
        return False

    if os.environ.get('CI') is None:
        log.error('annotate should only be used under CI context')
        return False

    if os.environ.get('CI_COMMIT_BRANCH') is None or os.environ['CI_COMMIT_BRANCH'] != 'master':
        log.error('annotate should only be used on master branch')
        return False

    if os.environ.get('GRAFANA_SERVER') is None:
        log.error('missing GRAFANA_SERVER envrionment variable')
        return False

    if os.environ.get('GRAFANA_USER') is None:
        log.error('missing GRAFANA_USER envrionment variable')
        return False

    if os.environ.get('GRAFANA_PASSWD') is None:
        log.error('missing GRAFANA_PASSWD envrionment variable')
        return False

    if os.environ.get('GITLAB_USER_LOGIN') is None:
        log.error('missing GITLAB_USER_LOGIN environment variable')
        return False

    if os.environ.get('CI_PIPELINE_URL') is None:
        log.error('missing CI_PIPELINE_URL environment variable')
        return False

    url = 'https://{user}:{passwd}@{server}/api/annotations'.format(
        server=os.environ['GRAFANA_SERVER'],
        user=os.environ['GRAFANA_USER'],
        passwd=os.environ['GRAFANA_PASSWD']
    )

    control = loadControl()
    build_number = control['build']
    if os.environ.get('CI_PIPELINE_IID') is not None:
        build_number = os.environ.get('CI_PIPELINE_IID')

    payload = {
        'time': int(time.time()) * 1000,
        'timeEnd': int(time.time()) * 1000,
        'tags': ['wapt', '{package}'.format(package=control['package'])],
        'text': '{package} {version}-{build} ({hash})\n'
                'builded by {user} - see <a href={url}>build {pipeline}</a>'.format(
            package=control['name'],
            version=control['version'],
            build=build_number,
            hash=os.environ['CI_COMMIT_SHORT_SHA'],
            user=os.environ['GITLAB_USER_LOGIN'],
            url=os.environ['CI_PIPELINE_URL'],
            pipeline=os.environ['CI_PIPELINE_ID'])}

    r = requests.post(url, json=payload)
    if r.status_code != 200:
        log.info('Annotation creation failed with code: {}'.format(r.status_code))

    return True
